# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
#
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create a plot of any hourly data by wind directions.
-

    Args:
        north_: An optional number between -360 and 360 for the counterclockwise
            difference between the North and the positive Y-axis in degrees. 90 is West
            and 270 is East. This can also be Vector for the direction to North
            (Default: 0).
        _data: A HourlyContinuousCollection or HourlyDiscontinuousCollection of
            values corresponding to the wind directions, which is "binned" by
            the direction intervals. This input usually consists of wind speed
            values, but is not limited to this data type. It can also be a list
            of data collections in which case multiple wind roses will be output.
        _wind_direction: A HourlyContinuousCollection or HourlyDiscontinuousCollection
            of wind directions which will be used to "bin" the _data items for the
            windrose.
        _dir_count_: Number that determines the number of directions to the wind rose
            will display. The number of directions must be greater then three to plot
            the wind rose (Default: 36).
        _center_pt_: Point3D to be used as a starting point to generate the geometry of
            the plot (Default: (0, 0, 0)).
        show_calm_: A boolean to indicate if the wind rose should display the fraction of
            time with zero wind speed using a circle in the center of the plot.
            The radius of this circle corresponds to the total amount of time
            with zero values divided by the number of directions. This means that
            the time period representing zero values is evenly distrobuted
            across all directions. (Default: False).
        show_avg_: A boolean to note whether the average value in each wind direction bin
            should be displayed instead of the complete frequency of _data
            values. (Default: False).
        _freq_dist_: The distance for the frequency interval in model units. If 
            show_calm_ is True, then the initial frequency interval corresponds
            to the number of calm hours in the data collection, which may not
            align with this freq_dist_ (Default: 5 meters)
        _freq_hours_: The number of hours in each frequency interval (Default: 50).
        _max_freq_lines_: A number representing the maximum frequency intervals in
            the rose, which determines the maximum amount of hours represented by the
            outermost ring of the windrose. Specifically, this number multiplied by the
            _freq_hours_ parameter will equal the maximum hours in that outermost
            ring. By default, this value is determined by the wind direction with the
            largest number of hours (the highest frequency) but you may want to change
            this if you have several wind roses that you want to compare to each other.
            For example, if you have wind roses for different months or seasons, which
            each have different maximum frequencies.
        legend_par_: An optional LegendParameter object to change the display of the
            WindRose plot. The number of segments in the legend determines the number of
            frequency intervals in the wind rose. If nothing is provided, a default
            LegendParameter object is computed using values from the wind data with
            11 segments (Default: None).
        statement_: A conditional statement as a string (e.g. a > 25) for
            the _data and _wind_direction inputs.
            .
            The variable of the first collection input to _data should always
            be named 'a' (without quotations), the variable of the second list
            should be named 'b', and so on. The wind direction is always the
            last variable, though most statements won't have a need for it.
            .
            For example, if three data collections are connected to _data
            and the following statement is applied:
            '18 < a < 26 and b < 80 and c > 2'
            The resulting collections will only include values where the first
            data collection is between 18 and 26, the second collection is less
            than 80 and the third collection is greater than 2.
        period_: An optional Ladybug analysis period to be applied to all of
            the input data.

    Returns:
        report: ...
        mesh: A colored mesh representing the wind rose derived from the input data.
            Multiple meshes will be output for several data collections are input.
        compass: A set of circles, lines and text objects that mark the cardinal
            directions in relation to the wind rose.
        orient_line: Line geometries representing the edges (or "spokes") of the wind
            rose directions.
        freq_line: Polygon geometries representing the frequency intervals of the wind
            rose.
        windrose_line: Polygon geometries representing the windrose outlines. This
            output is hidden by default and should be connected to a native
            Grasshopper Geometry component in order to be visualized.
        legend: Geometry representing the legend for the wind rose.
        title: A text object for the global_title.
        prevailing: The predominant direction of the outpt wind rose in clockwise
            degrees from north. 0 is North, 90 is East, 180 is South, 270 is West.
        angles: A list of angles corresponding to each windrose directions.
        calm_hours: The number of hours with calm wind speeds. Only returns a value
            if the input _data is wind speed. 
        histogram_data: The input _data in a histogram structure after it has gone
            through any of the statement or period operations input to
            this component.
        vis_set: An object containing VisualizationSet arguments for drawing a detailed
            version of the Wind Rose in the Rhino scene. This can be connected to
            the "LB Preview Visualization Set" component to display this version
            of the Wind Rose in Rhino.
"""

ghenv.Component.Name = 'LB Wind Rose'
ghenv.Component.NickName = 'WindRose'
ghenv.Component.Message = '1.8.1'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '2 :: Visualize Data'
ghenv.Component.AdditionalHelpFromDocStrings = '5'

import math

try:
    from ladybug.datacollection import HourlyContinuousCollection
    from ladybug.windrose import WindRose
    from ladybug.datatype.speed import Speed
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_geometry.geometry2d.pointvector import Point2D, Vector2D
    from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import conversion_to_meters
    from ladybug_rhino.togeometry import to_point3d, to_vector2d
    from ladybug_rhino.fromgeometry import from_mesh2d, from_linesegment2d, \
        from_polygon2d
    from ladybug_rhino.text import text_objects
    from ladybug_rhino.fromobjects import legend_objects, compass_objects
    from ladybug_rhino.grasshopper import all_required_inputs, list_to_data_tree, \
        objectify_output, hide_output
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def title_text(data_col):
    """Get a text string for the title of the windrose."""
    title_array = ['{} ({})'.format(data_col.header.data_type,
                                    data_col.header.unit)]
    for key, val in data_col.header.metadata.items():
        title_array.append('{}: {}'.format(key, val))
    title_array.append('period: {}'.format(data_col.header.analysis_period))
    return '\n'.join(title_array)


if all_required_inputs(ghenv.Component):
    # Apply any analysis periods and conditional statement to the input collections
    if period_ is not None:
        _data = [dat.filter_by_analysis_period(period_) for dat in _data]
        _wind_direction = _wind_direction.filter_by_analysis_period(period_)
    if statement_ is not None and statement_.strip() != "":
        _fdata = HourlyContinuousCollection.filter_collections_by_statement(
            _data + [_wind_direction], statement_)
        _data = _fdata[:-1]
        _wind_direction = _fdata[-1]

    # filter zero speed values out of collections if the speed is input
    pattern = []
    filt_wind_dir = _wind_direction
    for dat in _data:
        if isinstance(dat.header.data_type, Speed):
            for val in dat.values:
                pat = True if val > 1e-10 else False
                pattern.append(pat)
            break
    if len(pattern) != 0:
        for i, dat in enumerate(_data):
            if not isinstance(dat.header.data_type, Speed):
                _data[i] = dat.filter_by_pattern(pattern)
        filt_wind_dir = _wind_direction.filter_by_pattern(pattern)

    # check errors in dir_count and process the north input
    _dir_count_ = 36 if _dir_count_ is None else _dir_count_
    assert _dir_count_ > 2, 'The number of directions must be greater than 3 ' \
        'to plot the wind rose. Got: {}'.format(_dir_count_)
    if north_ is not None:  # process the north_
        try:
            north_ = math.degrees(to_vector2d(north_).angle_clockwise(Vector2D(0, 1)))
        except AttributeError:  # north angle instead of vector
            north_ = float(north_)
            assert -360.0 <= north_ <= 360.0, 'The north orientation must be greater ' \
                'then -360 and less then 360 to plot the wind rose. ' \
                'Got: {}'.format(north_)
    else:
        north_ = 0.0

    # set default values for the center point
    _center_pt_ = to_point3d(_center_pt_) if _center_pt_ is not None else Point3D()
    center_pt_2d = Point2D(_center_pt_.x, _center_pt_.y)

    # set defaults frequency hours and distance so chart is same scale as other LB plots
    if _freq_hours_ is None:
        _freq_hours_ = 50.0
    if _freq_dist_ is None:
        _freq_dist_ = 5.0 / conversion_to_meters()

    # set default show_freq and show_calm_
    show_calm_ = False if show_calm_ is None else show_calm_
    show_freq_ = True if show_avg_ is None else not show_avg_

    # set up empty lists of objects to be filled
    all_windrose_lines = []
    mesh = []
    all_compass = []
    all_orient_line = []
    all_freq_line = []
    all_legends = []
    title = []
    calm_hours = []
    histogram = []

    # Calculate _max_freq_lines_ if it's not already set, to use to
    # determine spacing for multiple plots.
    if len(_data) > 1 and _max_freq_lines_ is None:
        max_freqs = []
        for i, _data_item in enumerate(_data):
            win_dir = _wind_direction if isinstance(_data_item.header.data_type, Speed) \
                else filt_wind_dir
            w = WindRose(win_dir, _data_item, _dir_count_)
            w.frequency_hours = _freq_hours_
            w.frequency_spacing_distance = _freq_dist_
            max_freqs.append(w.frequency_intervals_compass)
        _max_freq_lines_ = max(max_freqs)

    # plot the windroses
    all_windroses = []
    for i, speed_data in enumerate(_data):
        # make the windrose
        win_dir = _wind_direction if isinstance(speed_data.header.data_type, Speed) \
            else filt_wind_dir
        windrose = WindRose(win_dir, speed_data, _dir_count_)
        all_windroses.append(windrose)

        # set the wind rose properties
        if len(legend_par_) > 0:
            try:  # sense when several legend parameters are connected
                lpar = legend_par_[i]
            except IndexError:
                lpar = legend_par_[-1]
            windrose.legend_parameters = lpar
        windrose.frequency_hours = _freq_hours_
        if _max_freq_lines_ is not None:
            windrose.frequency_intervals_compass = _max_freq_lines_
        windrose.frequency_spacing_distance = _freq_dist_
        windrose.north = north_
        windrose.show_freq = show_freq_

        calm_text = ''
        if isinstance(speed_data.header.data_type, Speed):
            windrose.show_zeros = show_calm_
            calm_text = '\nCalm for {}% of the time = {} hours.'.format(
                round(windrose._zero_count / 
                len(windrose.analysis_values) * 100.0, 2),
                windrose._zero_count)
        windrose.base_point = Point2D(center_pt_2d.x, center_pt_2d.y)

        # Make the mesh
        msh = from_mesh2d(windrose.colored_mesh, _center_pt_.z)

        # Make the other graphic outputs
        lb_legend = windrose.legend
        ttl_pt = windrose.container.lower_title_location
        if _center_pt_.z != 0:
            move_vec = Vector3D(0, 0, _center_pt_.z)
            ttl_pt = ttl_pt.move(move_vec)
            if lb_legend.legend_parameters.is_base_plane_default:
                lb_legend = lb_legend.duplicate()
                lb_legend.legend_parameters.base_plane = \
                    lb_legend.legend_parameters.base_plane.move(move_vec)
        legend = legend_objects(lb_legend)
        freq_per = windrose._frequency_hours / \
            len([b for a in windrose.histogram_data for b in a])
        freq_text = '\nEach closed polyline shows frequency of {}% = {} hours.'.format(
                round(freq_per * 100, 1), windrose._frequency_hours)
        titl = text_objects(title_text(speed_data) + calm_text + freq_text, ttl_pt,
                            windrose.legend_parameters.text_height,
                            windrose.legend_parameters.font)
        compass = compass_objects(windrose.compass, _center_pt_.z, None)
        orient_line = [from_linesegment2d(seg, _center_pt_.z)
                       for seg in windrose.orientation_lines]
        freq_line = [from_polygon2d(poly, _center_pt_.z) for poly in windrose.frequency_lines]
        windrose_lines = [from_polygon2d(poly, _center_pt_.z) for poly in windrose.windrose_lines]
        fac = (i + 1) * windrose.compass_radius * 3
        center_pt_2d = Point2D(_center_pt_.x + fac, _center_pt_.y)

        # collect everything to be output
        mesh.append(msh)
        all_compass.append(compass)
        all_orient_line.append(orient_line)
        all_freq_line.append(freq_line)
        all_windrose_lines.append(windrose_lines)
        all_legends.append(legend)
        title.append(titl)
        calm = windrose.zero_count if isinstance(speed_data.header.data_type, Speed) else None
        calm_hours.append(calm)
        histogram.append(objectify_output('WindRose {}'.format(i), windrose.histogram_data))

    # convert nested lists into data trees
    compass = list_to_data_tree(all_compass)
    orient_line = list_to_data_tree(all_orient_line)
    freq_line = list_to_data_tree(all_freq_line)
    windrose_line = list_to_data_tree(all_windrose_lines)
    legend = list_to_data_tree(all_legends)
    hide_output(ghenv.Component, 5)  # keep the devault visual simple

    # compute direction angles and prevailing direction
    theta = 180.0 / _dir_count_
    angles = [(angle + theta) % 360.0 for angle in windrose.angles[:-1]]
    prevailing = windrose.prevailing_direction
    vis_set = []
    for wr in all_windroses:
        vis_set.append(objectify_output('VisualizationSet Aruments [WindRose]', [wr, _center_pt_.z]))



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Import location from an epw file.
-

    Args:
        _epw_file: An epw file path on your system as a string.
    
    Returns:
        location: Location data (use this output to construct the sun path).
"""

ghenv.Component.Name = 'LB Import Location'
ghenv.Component.NickName = 'ImportLoc'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '0 :: Import'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

try:
    import ladybug.epw as epw
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    ep = epw.EPW(_epw_file)
    location = ep.location


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Compute solar infomation about a day of the year at a particular location. This includes
the time of sunrise, sunset, solar noon, and the length of the day in hours.
_
Note that these times are intended to represent a typical year and they will often
vary by a few minutes depending on where in the leap year cycle a given year falls.
-

    Args:
        _location: A ladybug Location that has been output from the "LB Import EPW"
            component or the "LB Construct Location" component.
        _doy: An integer for the day of the year for which solar information is
            be computed. The "LB Calculate HOY" component can be used to compute
            the day of the year from month and day inputs.
        _depression_: An angle in degrees indicating the additional period before/after
            the edge of the sun has passed the horizon where the sun is still
            considered up. Setting this value to 0 will compute sunrise/sunset
            as the time when the edge of the sun begins to touch the horizon.
            Setting it to the angular diameter of the sun (0.5334) will compute
            sunrise/sunset as the time when the sun just finishes passing the
            horizon (actual physical sunset). Setting it to 0.833 will compute
            the apparent sunrise/sunset, accounting for atmospheric refraction.
            Setting this to 6 will compute sunrise/sunset as the beginning/end
            of civil twilight. Setting this to 12 will compute sunrise/sunset
            as the beginning/end of nautical twilight. Setting this to 18 will
            compute sunrise/sunset as the beginning/end of astronomical
            twilight. (Default: 0.5334 for the physical sunset).
        solar_time_: A boolean to indicate if the output datetimes for sunrise,
            noon and sunset should be in solar time as opposed to the time zone
            of the location. (Default: False).
        dl_saving_: An optional analysis period for daylight saving time. This will be used
            to adjust the output times by an hour when applicable. If unspecified,
            no daylight saving time will be used

    Returns:
        sunrise: The time of sunrise expressed as HH:MM where hours range from 0 to 23.
            Note that this may be None if there is no sunrise or sunset on the
            specified day. (eg. at the north pole on the winter solstice).
        sunset: The time of sunset expressed as HH:MM where hours range from 0 to 23.
            Note that this may be None if there is no sunrise or sunset on the
            specified day. (eg. at the north pole on the winter solstice).
        solar_noon: The time of solar noon when the sun is at its highest point in the
            sky, expressed as HH:MM.
        noon_alt: The altitude of the sun at solar noon in degrees. This is the maximum
            altitude that will be expereinced on the input day.
        day_length: The length of the input day in hours.
"""

ghenv.Component.Name = 'LB Day Solar Information'
ghenv.Component.NickName = 'DayInfo'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug.sunpath import Sunpath
    from ladybug.dt import Date
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # set default values
    solar_time_ = False if solar_time_ is None else solar_time_  # process solar time
    _depression_ = 0.5334 if _depression_ is None else _depression_

    # initialize sunpath based on location
    sp = Sunpath.from_location(_location, 0, dl_saving_)

    # for each day of the year, compute the information
    sunrise, sunset, solar_noon, noon_alt, day_length = [], [], [], [], []
    for doy in _doy:
        doy_date = Date.from_doy(doy)
        solar_info = sp.calculate_sunrise_sunset(
            doy_date.month, doy_date.day, _depression_, solar_time_)
        print(solar_info)
        sr, sn, ss = solar_info['sunrise'], solar_info['noon'], solar_info['sunset']
        solar_noon.append(sn.time)
        noon_alt.append(sp.calculate_sun_from_date_time(sn).altitude)
        if sr is not None:
            sunrise.append(sr.time)
        else:
            sunrise.append(None)
        if ss is not None:
            sunset.append(ss.time)
            day_length.append((ss - sr).total_seconds() / 3600)
        else:
            sunset.append(None)
            day_length.append(None)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Calculate view factors from a point or plane to a set of geometries.
_
View factors are used in many thermal comfort calculations such as mean radiant
temperture (MRT) or discomfort from radiant assymetry. 
-

    Args:
        _study_point: A point or plane from which view vectors will be projected.
            Note that, if a point is connected, all view vectors will be
            weighted evenly (assuming no directional bias). However, if
            a plane is connected, vectors will be weighted based on their
            angle to the plane normal, producing view factors for a surface
            in the connected plane. The first is useful for MRT calculations
            while the latter is needed for radiant assymetry calculations.
            This input can also be a list of several points or planes.
        _view_geo: A list of breps, surfaces, or meshes to which you want to compute
            view factors. Note that by meshing and joining several goemtries
            together, the combined view factor to these geometries can
            be computed.
        context_: Optional context geometry as breps, surfaces, or meshes that can
            block the view to the _view_geo.
        _resolution_: A positive integer for the number of times that the original
            view vectors are subdivided. 1 indicates that 145 evenly-spaced
            vectors are used to describe a hemisphere, 2 indicates that
            577 vectors describe a hemisphere, and each successive value will
            roughly quadruple the number of view vectors used. Setting this to
            a high value will result in a more accurate analysis but will take
            longer to run. (Default: 1).
        _cpu_count_: An integer to set the number of CPUs used in the execution of the
            intersection calculation. If unspecified, it will automatically default
            to one less than the number of CPUs currently available on the
            machine or 1 if only one processor is available.
        _run: Set to True to run the component and claculate view factors.

    Returns:
        report: ...
        view_vecs: A list of vectors which are projected from each of the points to evaluate view.
        patch_mesh: A mesh that represents the sphere of view patches around the _study_point
            at the input _resolution_. There is one face per patch and this can be
            used along with the int_mtx to create a colored visualization of patches
            corresponding to different geometries around the point. Specifically,
            the "LB Spaital Heatmap" component is recommended for such visualizations.
            Note that only one sphere is ever output from here and, in the event
            that several _study_points are connected, this sphere will be located
            at the first point. Therefore, to create visualizations for the other
            points, this mesh should be moved using the difference between the
            first study point and following study points.
        view_factors: A list of view factors that describe the fraction of sperical
            view taken up by the input surfaces.  These values range from 0 (no view)
            to 1 (full view).  If multiple _study_points have been connected,
            this output will be a data tree with one list for each point.
        int_mtx: A Matrix object that can be connected to the "LB Deconstruct Matrix"
            component to obtain detailed vector-by-vector results of the study.
            Each sub-list (aka. branch of the Data Tree) represents one of the
            points used for analysis. Each value in this sub-list corresponds
            to a vector used in the study and the value denotes the index of the
            geometry that each view vector hit. This can be used to identify
            which view pathces are intersected by each geometry. If no geometry
            is intersected by a given vector, the value will be -1.
"""

ghenv.Component.Name = 'LB View Factors'
ghenv.Component.NickName = 'ViewFactors'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '3 :: Analyze Geometry'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

try:
    from ladybug.viewsphere import view_sphere
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.togeometry import to_plane, to_vector3d
    from ladybug_rhino.fromgeometry import from_mesh3d, from_point3d, from_vector3d
    from ladybug_rhino.intersect import join_geometry_to_mesh, intersect_view_factor
    from ladybug_rhino.grasshopper import all_required_inputs, hide_output, \
        show_output, objectify_output, list_to_data_tree, recommended_processor_count
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _run:
    # set up the defaults
    _resolution_ = _resolution_ if _resolution_ is not None else 1
    workers = _cpu_count_ if _cpu_count_ is not None else recommended_processor_count()

    # process the input points and determine whether they are points or planes
    points, normals = [], []
    for geo in _study_point:
        try:
            test_plane = to_plane(geo)
            points.append(from_point3d(test_plane.o))
            normals.append(from_vector3d(test_plane.n))
        except AttributeError:  # it is a point
            points.append(geo)
            normals.append(None)
    if all(n is None for n in normals):
        normals = None  # none of the inputs were planes

    # generate the view vectors based on the resolution
    patch_mesh, lb_vecs = view_sphere.sphere_patches(_resolution_)
    patch_wghts = view_sphere.sphere_patch_weights(_resolution_)
    # correct for the fact that the last patch has several mesh faces
    patch_count = 144 * (_resolution_ ** 2) + 1
    extend_count = ((6 * _resolution_) - 1)
    up_dome, down_dome = list(lb_vecs[:patch_count]), list(lb_vecs[patch_count:])
    up_dome.extend([up_dome[-1]] * extend_count)
    down_dome.extend([down_dome[-1]] * extend_count)
    lb_vecs = up_dome + down_dome
    up_weights, down_weights = list(patch_wghts[:patch_count]), list(patch_wghts[patch_count:])
    up_weights.extend([up_weights[-1] / extend_count] * extend_count)
    down_weights.extend([down_weights[-1] / extend_count] * extend_count)
    patch_wghts = up_weights + down_weights

    # process the context if it is input
    context_mesh = join_geometry_to_mesh(context_) if len(context_) != 0 else None

    # perform the intersection to compute view factors
    view_vecs = [from_vector3d(pt) for pt in lb_vecs]
    view_factors, mesh_indices = intersect_view_factor(
        _view_geo, points, view_vecs, patch_wghts,
        context_mesh, normals, cpu_count=workers)

    # convert the outputs into the correct Grasshopper format
    view_factors = list_to_data_tree(view_factors)
    int_mtx = objectify_output('View Factor Intersection Matrix', mesh_indices)
    patch_mesh = from_mesh3d(patch_mesh.move(to_vector3d(points[0])))
    hide_output(ghenv.Component, 2)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Dragonfly.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Dragonfly; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
This component updates the Ladybug Tools core libraries and grasshopper components
to either the latest development version available (default) or to a specific
version of the grasshopper plugin.
_
The input version_ does not need to be newer than the current installation and can
be older but grasshopper plugin versions less than 0.3.0 are not supported.
A list of all versions of the Grasshopper plugin and corresponding release notes
can be found at: https://github.com/ladybug-tools/lbt-grasshopper/releases
-
    Args:
        _update: Set to True to update your installation of Ladybug Tools to the
            latest development version or to be at the version specified below.
        version_: An optional text string for the version of the grasshopper
            plugin which you would like to update to. The input should contain
            only integers separated by two periods (eg. 1.0.0). The version does
            not need to be newer than the current installation and can be older
            but grasshopper plugin versions less than 0.3.0 are not supported.
            A list of all versions of the Grasshopper plugin can be found
            here - https://github.com/ladybug-tools/lbt-grasshopper/releases

    Returns:
        Vviiiiiz!: !!!
"""

ghenv.Component.Name = 'LB Versioner'
ghenv.Component.NickName = 'Versioner'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '5 :: Version'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

import os
import tempfile
import subprocess

try:
    from ladybug.config import folders
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.versioning.diff import current_userobject_version
    from ladybug_rhino.versioning.change import latest_github_version
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning, \
        give_popup_message, is_user_admin
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _update:
    # get the paths to the executables
    lb_rhino_exe = os.path.join(folders.python_scripts_path, 'ladybug-rhino.exe') \
        if os.name == 'nt' else os.path.join(folders.python_scripts_path, 'ladybug-rhino')
    executor_path = os.path.join(
        folders.ladybug_tools_folder, 'grasshopper', 'ladybug_grasshopper_dotnet',
        'Ladybug.Executor.exe')

    # run the command to update everything
    if os.name == 'nt' and os.path.isfile(executor_path) and \
            'Program Files' in executor_path and not is_user_admin():
        cmd = [
            executor_path, folders.python_exe_path,
            '-m ladybug_rhino change-installed-version'
        ]
    else:
        cmd = [lb_rhino_exe, 'change-installed-version']
    if version_ is not None:
        cmd.extend(['--version', version_])
    use_shell = True if os.name == 'nt' else False
    custom_env = os.environ.copy()
    custom_env['PYTHONHOME'] = ''
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=use_shell, env=custom_env)
    cmd_output = process.communicate()  # ensure that the canvas is frozen
    stdout = cmd_output[0]
    print(stdout)
    if 'Error' in stdout:
        if 'ladybug-rhino.exe' not in stdout:
            raise ValueError(stdout)
        else:  # recent Windows permission issue; reinstall ladybug-rhino
            cmd = cmds = [folders.python_exe_path, '-m', 'pip', 'install', 'ladybug-rhino', '-U']
            process = subprocess.Popen(
                cmd, stdout=subprocess.PIPE, shell=use_shell, env=custom_env)
            cmd_output = process.communicate()  # ensure that the canvas is frozen

    # give a completion message
    if version_ is None:
        temp_dir = tempfile.gettempdir()
        version_ = latest_github_version('lbt-grasshopper', temp_dir)
    version = 'LATEST' if version_ is None else version_
    success_msg = 'Change to Version {} Successful!'.format(version)
    restart_msg = 'RESTART RHINO to load the new components + library.'
    sync_msg = 'The "LB Sync Grasshopper File" component can be used\n' \
        'to sync Grasshopper definitions with your new installation.'
    for msg in (success_msg, restart_msg, sync_msg):
        print (msg)
    give_popup_message('\n'.join([restart_msg, sync_msg]), success_msg)

    # do a check to see if the versioner has, itself, been updated
    new_version = current_userobject_version(ghenv.Component)
    current_version = ghenv.Component.Message
    if new_version != current_version:
        msg = 'The Versioner component has, itself, been changed between the\n' \
            'current version ({}) and the version you are changing to ({}).\n' \
            'It is recommended that you resart Rhino and run the new Versioner\n' \
            'coponent to ensure that everything is consistent.'.format(
                current_version, new_version)
        print (msg)
        give_warning(ghenv.Component, msg)
else:  # give a message to the user about what to do
    print ('Make sure you are connected to the internet and set _update to True!')



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Export a Ladybug Tools Grasshopper GHPython component as a UserObject that can
be installed on other's machines.
-

    Args:
        _components: A Ladybug Tools GHPython component to be exported. This
            can also be a list of of components to be exported together. Lastly,
            this can be a "*" and all of the Ladybug Tools components on the
            Grasshopper canvass will be exported.
        _folder: Full path to folder to copy the updated UserObject and the
            source code. It is usually path to where you have cloned the
            repository from GitHub. Exported contents will be created under src
            and userobject folders.
        _change_type_: One of the values listed below based on the type of
            change made to the component. Export component will validate the
            version of the newly created UserObject against the version of the
            current installed UserObject with the same name based on the change
            type. If an older version of the component does not exist the
            current version of the component will be used. The version is
            structured as major.minor.patch. (Default: fix)
            _
            * release: You are changing the versions for a new release.
                Bump the major with 1 and set minor and patch to 0.
            _
            * feat: You have added a new feature. Adding a new feature usually
                results in a change in inputs or outputs of the component.
                Bump minor by 1 and set patch to 0.
            _
            * perf: You have improved the component for better performance.
                Similar to adding a feature you should bump the minor by 1 and
                set patch to 0.
            _
            * fix: You have fixed the code inside the component. It results in
                a single bump in patch.
            _
            * docs: You have improved the documentation. No change in version.
            _
            * ignore: This is an exception to the rule and you want the change
                type to be ignored. You should use this option only in rare
                occasional cases.
        _export: Set to True to export the component.

    Returns:
        report: Errors, warnings, etc.
"""

ghenv.Component.Name = 'LB Export UserObject'
ghenv.Component.NickName = 'ExportUO'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '5 :: Version'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug_rhino.versioning.gather import gather_canvas_components, \
        gather_connected_components
    from ladybug_rhino.versioning.diff import validate_change_type
    from ladybug_rhino.versioning.export import export_component, refresh_toolbar
    from ladybug_rhino.grasshopper import turn_off_old_tag
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))
turn_off_old_tag(ghenv.Component)  # turn off the OLD tag in Rhino 8


if _export and _folder and len(_components) != 0:
    change_type = validate_change_type(_change_type_)
    comps_to_export = gather_connected_components(ghenv.Component) \
        if str(_components[0]) != '*' \
        else gather_canvas_components(ghenv.Component)
    for comp in comps_to_export:
        print('Processing %s...' % comp.Name)
        export_component(_folder, comp, _change_type_)
    refresh_toolbar()  # try to update the toolbar
else:
    print('Connect _components and _folder')



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Visualize the view openness around a given point as a colored mesh that fills a
circle, sphere, or hemisphere (depending on the specified view type).
_
The input context will block the view, resulting in a decresed view rose size
and a change in the view rose color.
-

    Args:
        _context: Rhino Breps or Meshes representing context geometry that can block
            the view around the center point.
        _center_pt: A point for the center of the view rose from which view openness
            will be evaluated.
        _radius_: A number for the radius of the view rose in Rhino model units. This is
            also used to evaluate the distance at which context is no longer
            able to block the view from the center point. This value should
            typically be increased if the view rose does not extend past the
            _context geometry. (Default: 100 meters in the current Rhino model
            units system).
        _view_type_: Text or an integer representing the type of view analysis to conduct.
            Choose from the following options. (Default: 0 - HorizontalRadial)
                _
                0 - HorizontalRadial - The percentage of the 360 horizontal view
                    plane that is not blocked by the context geometry.
                _
                1 - Horizontal30DegreeOffset - The percentage of the 360 horizontal
                    view band bounded on top and bottom by a 30 degree offset from
                    the horizontal plane. 30 degrees corresponds roughly to the
                    vertical limit of human peripheral vision.
                _
                2 - Spherical - The percentage of the sphere surrounding each of
                    the test points that is not blocked by context geometry. This
                    is equivalent to a solid angle and gives equal weight to all
                    portions of the sphere.
                _
                3 - SkyExposure - The percentage of the sky that is visible from
                    each of the the test points.
        _resolution_: A positive integer for the number of times that the original
            view vectors are subdivided. For a circle, 1 indicates that 72
            evenly-spaced vectors are used to describe a circle, 2 indicates
            that 144 vectors describe a circle, and each successive value will
            roughly double the number of view vectors used. For a dome, 1 indicates
            that 1225 are used to describe the dome, 2 indicates that 5040
            view vectors describe the some and each successive value will
            roughly quadruple the number of view vectors used. Setting this to
            a high value will result in a more accurate analysis but will take
            longer to run. (Default: 1).
        legend_par_: Optional legend parameters from the "LB Legend Parameters"
            that will be used to customize the display of the results.


    Returns:
        report: ...
        view_vecs: A list of vectors which are projected from each of the points
            to evaluate view.
        results: A list of numbers that aligns with the vertices of the mesh. Each number
            indicates the distance from the _center_pt at which the view is
            blocked from a particular direction.
        mesh: A colored mesh representing the visible area from the viewpoint past
            the _context geometry. Colors indicate how open the view is from a
            given direction.
        legend: A legend that correspond to the colors of the mesh and shows the distance
            at which vectors are blocked.
"""

ghenv.Component.Name = 'LB View Rose'
ghenv.Component.NickName = 'ViewRose'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '3 :: Analyze Geometry'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

try:
    from ladybug_geometry.geometry3d import Mesh3D
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:
    from ladybug.viewsphere import view_sphere
    from ladybug.color import Colorset
    from ladybug.graphic import GraphicContainer
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import conversion_to_meters, units_abbreviation
    from ladybug_rhino.togeometry import to_point3d
    from ladybug_rhino.fromgeometry import from_mesh3d, from_vector3d
    from ladybug_rhino.fromobjects import legend_objects
    from ladybug_rhino.intersect import join_geometry_to_mesh, \
        intersect_mesh_rays_distance
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


# dictionary to record all available view types
VIEW_TYPES = {
    None: (0, view_sphere.horizontal_circle_view_mesh),
    'HorizontalRadial': (0, view_sphere.horizontal_circle_view_mesh),
    'Horizontal30DegreeOffset': (1, view_sphere.horizontal_radial_view_mesh),
    'Spherical': (2, view_sphere.sphere_view_mesh),
    'SkyExposure': (3, view_sphere.dome_view_mesh),
    '0': (0, view_sphere.horizontal_circle_view_mesh),
    '1': (1, view_sphere.horizontal_radial_view_mesh),
    '2': (2, view_sphere.sphere_view_mesh),
    '3': (3, view_sphere.dome_view_mesh)
}


if all_required_inputs(ghenv.Component):
    # process the view_type_ and set the default values
    vt_int, view_method = VIEW_TYPES[_view_type_]
    _radius_ = int(100 / conversion_to_meters()) if _radius_ is None else _radius_
    resolution_ = _resolution_ if _resolution_ is not None else 1
    az_count = 72 * resolution_
    alt_count = 6 * resolution_ if vt_int == 1 else 18 * resolution_

    # get the view vectors and mesh based on the inputs
    lb_point = to_point3d(_center_pt)
    if vt_int == 0:
        study_mesh, lb_vecs = view_method(
            center_point=lb_point, radius=_radius_, azimuth_count=az_count)
    else:
        study_mesh, lb_vecs = view_method(
            center_point=lb_point, radius=_radius_,
            azimuth_count=az_count, altitude_count=alt_count)
    view_vecs = [from_vector3d(pt) for pt in lb_vecs]

    # process the input geometry
    shade_mesh = join_geometry_to_mesh(_context)

    # intersect the rays with the mesh
    results = intersect_mesh_rays_distance(shade_mesh, _center_pt, view_vecs, _radius_)

    # adjust the mesh based on the distance
    move_vecs = [vec * -(_radius_ - dist) for vec, dist, in zip(lb_vecs, results)]
    new_verts = [lb_point] if vt_int in (0, 1) else []
    iter_verts = study_mesh.vertices[1:] if  vt_int in (0, 1) else study_mesh.vertices
    for pt, mv in zip(iter_verts, move_vecs):
        new_verts.append(pt.move(mv))
    study_mesh = Mesh3D(new_verts, study_mesh.faces)

    # add a value at the start to align with the vertices
    if vt_int in (0, 1):
        avg_val = sum(results) / len(results)
        results.insert(0, avg_val)

    # create the mesh and legend outputs
    graphic = GraphicContainer(results, study_mesh.min, study_mesh.max, legend_par_)
    graphic.legend_parameters.title = units_abbreviation()
    if legend_par_ is None or legend_par_.are_colors_default:
        graphic.legend_parameters.colors = Colorset.view_study()

    # create all of the visual outputs
    study_mesh.colors = graphic.value_colors
    mesh = from_mesh3d(study_mesh)
    legend = legend_objects(graphic.legend)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Calculate the percentage of people dissatisfied from cold drafts at ankle-level.
_
The original tests used to create the model involved blowing cold air on
subject's ankles at a height of 10 cm off of the ground.
The formula was officially incorporated in the ASHRAE 55 standard in 2020
with a recommendation that PPD from ankle draft not exceed 20%.
_
For more information on the methods used to create this model see the following:
Liu, S., S. Schiavon, A. Kabanshi, W. Nazaroff. 2016. "Predicted
percentage of dissatisfied with ankle draft." Accepted Author Manuscript.
Indoor Environmental Quality. http://escholarship.org/uc/item/9076254n
-

    Args:
        _full_body_pmv: The full-body predicted mean vote (PMV) of the subject.
            Ankle draft depends on full-body PMV because subjects are more likely
            to feel uncomfortably cold at their extremities when their whole body
            is already feeling colder than neutral. The "LB PMV Comfort"
            component can be used to obatin this input here.
        _draft_velocity: The velocity of the draft in m/s at ankle level
            (10cm above the floor).

    Returns:
        ppd: The percentage of people dissatisfied (PPD) from cold drafts at ankle level.
"""

ghenv.Component.Name = 'LB Ankle Draft'
ghenv.Component.NickName = 'AnkleDraft'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug_comfort.local import ankle_draft_ppd
    from ladybug.datacollection import HourlyContinuousCollection
    from ladybug.datatype.fraction import PercentagePeopleDissatisfied
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    ppd = HourlyContinuousCollection.compute_function_aligned(
        ankle_draft_ppd, [_full_body_pmv, _draft_velocity],
        PercentagePeopleDissatisfied(), '%')



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Filter a data collection or list of data collections by a pattern of True/False
values. True values will be kept while False values will be filtered out.
-

    Args:
        _data: A Data Collection or list of aligned Data Collections to be filtered
            by a pattern.
        _pattern: A list of True/False values. Typically, list has a length matching
            the length of the Data Collection(s)'s values. However, it can also
            be a pattern to be repeated over the Data Collection(s)'s values.

    Returns:
        data: A list of Data Collections that have been filtered by the _pattern.
"""

ghenv.Component.Name = 'LB Apply Pattern'
ghenv.Component.NickName = 'Pattern'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug.datacollection import BaseCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    data = []
    for dat in _data:
        assert isinstance(dat, BaseCollection), '_data must be a data' \
            ' collection. Got {}.'.format(type(dat))
        data.append(dat.filter_by_pattern(_pattern))


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Construct a Ladybug data collection from header and values.
-

    Args:
        _header:A Ladybug header object describing the metadata of the data collection.
        _values: A list of numerical values for the data collection.
        _interval_: Text to indicate the time interval of the data collection, which
            determines the type of collection that is output. (Default: hourly).
            _
            Choose from the following:
                - hourly
                - daily
                - monthly
                - monthly-per-hour
            _
            Note that the "hourly" input is also used to represent sub-hourly
            intervals (in this case, the timestep of the analysis period
            must not be 1).

    Returns:
        data: A Ladybug data collection object.
"""

ghenv.Component.Name = "LB Construct Data"
ghenv.Component.NickName = '+Data'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from ladybug.datacollection import HourlyContinuousCollection, DailyCollection, \
        MonthlyCollection, MonthlyPerHourCollection, HourlyDiscontinuousCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    inter = _interval_.lower() if _interval_ is not None else 'hourly'
    if inter == 'hourly':
        aper = _header.analysis_period
        if aper.st_hour == 0 and aper.end_hour == 23:
            data = HourlyContinuousCollection(_header, _values)
        else:
            data = HourlyDiscontinuousCollection(_header, _values, aper.datetimes)
    elif inter == 'monthly':
        data = MonthlyCollection(
            _header, _values, _header.analysis_period.months_int)
    elif inter == 'daily':
        data = DailyCollection(
            _header, _values, _header.analysis_period.doys_int)
    elif inter == 'monthly-per-hour':
        data = MonthlyPerHourCollection(
            _header, _values, _header.analysis_period.months_per_hour)
    else:
        raise ValueError('{} is not a recongized interval.'.format(_interval_))



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Visualize the desirability of shade in terms of proximity of conditions to
a favorable temerature range.
_
The calculation runs by generating solar vectors for a data collection of input
temperature values. Solar vectors for hours when the temperature is above the
upper temperature threshold contribute positively to shade desirability (shade_help)
while solar vectors for hours when the temperature is below the lower temperature
threshold contribute negatively (shade_harm).
_
The component outputs a colored mesh of the shade illustrating the net effect of
shading each part of the _shade_geo. A higher saturation of blue indicates that
shading the cell is desirable to avoid excessively hot temperatures. A higher
saturation of red indicates that shading the cell is harmful, blocking helpful
sun in cold conditions that could bring conditions closer to the desired
temperature range. Desaturated cells indicate that shading the cell will have
relatively little effect on keeping the _study_region in the desired thermal range.
_
The units for shade desirability are degree-days per unit area of shade, which are
essentially the amount of time in days that sun is blocked by a given cell
multiplied by the degrees above (or below) the temperature thresholds during
that time. So, if a given square meter of input _shade_geo has a shade desirability
of 10 degree-days per square meter, this means that a shade in this location
provides roughly 1 day of sun protection from conditions 10 degrees Celsius
warmer than the _up_threshold_ to the _study_region.
_
More information on the methods used by this component can be found in the
following publication:
Mackey, Christopher; Sadeghipour Roudsari, Mostapha; Samaras, Panagiotis.
“ComfortCover: A Novel Method for the Design of Outdoor Shades.” In Proceedings
of Symposium on Simulation for Architecture and Urban Design. Washington, DC,
United States, Apr 12-15 2015.
https://drive.google.com/file/d/0Bz2PwDvkjovJQVRTRHhMSXZWZjQ/view?usp=sharing
-

    Args:
        north_: A number between -360 and 360 for the counterclockwise
            difference between the North and the positive Y-axis in degrees.
            90 is West and 270 is East. This can also be Vector for the
            direction to North. (Default: 0)
        _location: A ladybug Location that has been output from the "LB Import EPW"
            component or the "LB Construct Location" component.
        _temperature: An hourly data collection with the unshaded temperature experienced
            at the _study_region. This temperature will be used to evaluate shade
            benefit for this study region. This temperature data collection should
            typically be informed by an analysis with the "LB Outdoor Solar MRT"
            or the "LB Indoor Solar MRT" component, which will account for the
            increased temperature delta expereinced as a result of being in the
            sun. For evaluation of shade in terms of outdoor thermal comfort, the
            best practice is to use the Universal Thermal Climate Index (UTCI)
            temperature at the study region for this input. For evaluation of
            shade benefit in terms of indoor comfort,  the best practice is to
            use the Standard Effective Temperature (SET) derived from the
            "LB PMV Comfort" component for this input. In both cases, the MRT
            inputs to the thermal comfort models should use solar-adjusted MRT.
        _study_region: Rhino Breps and/or Rhino Meshes representing an area for which shading
            desirability is being evaluated. This is often the region where a human
            subject will sit (eg. a bench) or it could be the window of a building
            where an occupant might be standing or sitting.
        _shade_geo: Rhino Breps and/or Rhino Meshes representing shading to be evaluated
            in terms of its benefit. Note that, in the case that multiple
            shading geometries are connected, this component does not account
            for the interaction between the different shading surfaces and will
            just evaluate each part of the shade independently.
        context_: Rhino Breps and/or Rhino Meshes representing context geometry that can
            block sunlight to the _study_region, therefore discounting any benefit
            or harm that could come to the region.
        _grid_size: A positive number in Rhino model units for the size of grid cells at
            which the input _shade_geo will be subdivided for shade benefit
            analysis. The smaller the grid size, the higher the resolution of
            the analysis and the longer the calculation will take.  So it is
            recommended that one start with a large value here and decrease
            the value as needed. However, the grid size should usually be
            smaller than the dimensions of the smallest piece of the _shade_geo
            and context_ in order to yield meaningful results.
        _up_threshold_: A number representing the temperature in Celsius above which
            shade is considered desirable/helpful. The default is 26C, which
            corresponds to the upper limit of "No Thermal Stress" according
            to the UTCI thermal comfort model (above this, heat stress begins).
            A different value may be desirable for indoor thermal comfort studies.
        _low_threshold_: A number representing the temperature in Celsius below which shade
            is considered harmful and access to the sun is preferable. The default
            is 9C, which corresponds to the lower limit of "No Thermal Stress"
            according to the UTCI thermal comfort model (below this, cold stress
            begins). A different value may be desirable for indoor thermal
            comfort studies.
        _timestep_: An integer for the number of timesteps per hour at which
            sun vectors will be generated for the analysis. Higher values will
            result in the generation of more vectors, which will make the
            resulting shade mesh smoother and a better representation of shade
            benefit and harm. However, the calculation will take
            longer as there are more intersection operations to perform. The
            default is 1 timestep per hour, which is the coarsest resolution
            avalable and the fastest calculation.
        legend_par_: Optional legend parameters from the "LB Legend Parameters"
            that will be used to customize the display of the results.
        _cpu_count_: An integer to set the number of CPUs used in the execution of the
            intersection calculation. If unspecified, it will automatically default
            to one less than the number of CPUs currently available on the
            machine or 1 if only one processor is available.
        _run: Set to "True" to run the component and perform shade benefit analysis.

    Returns:
        report: ...
        vectors: The sun vectors that were used to evaluate the shade (note that
            these will increase as the _timestep_ increases).
        points: Points across the study_region from which sun vectors are projected.
        mesh: A colored mesh of the _shade_geo showing where shading is helpful (in blue),
            harmful (in red), or does not make much of a difference (white or
            desaturated colors). Note that the colors can change depending upon
            the input legend_par_.
        legend: Legend showing the numeric values of degree-days per unit are of shade
            that correspond to the colors in the shade mesh.
        title: A text object for the study title.
        shade_help: The cumulative degree-days per square area unit helped by shading each
            cell of the shade. If a given square meter of _shade_geo has a helpfulness
            of 10 degree-days/m2, this means that a shade in this location provides
            1 day of sun protection from conditions 10 degrees warmer than the
            _up_threshold_ to the _study_region.
        shade_harm: The cumulative degree-days per square area unit harmed by shading each
            cell of the shade. If a given square meter of _shade_geo has a harmfulness
            of -10 degree-days, this means that a shade in this location blocks
            1 day of sun duirng conditions that are 10 degrees Celsius colder than
            the _low_threshold_ to the _study_region.
        shade_net: The sum of the helpfulness and harmfulness for each cell. This will be
            negative if shading the cell has a net harmful effect and positive
            if the shade has a net helpful effect.
"""

ghenv.Component.Name = 'LB Thermal Shade Benefit'
ghenv.Component.NickName = 'ThermalShadeBenefit'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '3 :: Analyze Geometry'
ghenv.Component.AdditionalHelpFromDocStrings = '4'

import math

try:
    from ladybug.sunpath import Sunpath
    from ladybug.datacollection import HourlyDiscontinuousCollection
    from ladybug.color import Colorset
    from ladybug.graphic import GraphicContainer
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import units_abbreviation
    from ladybug_rhino.togeometry import to_joined_gridded_mesh3d, to_vector2d
    from ladybug_rhino.fromgeometry import from_mesh3d, from_point3d, from_vector3d
    from ladybug_rhino.fromobjects import legend_objects
    from ladybug_rhino.text import text_objects
    from ladybug_rhino.intersect import join_geometry_to_mesh, generate_intersection_rays, \
        intersect_rays_with_mesh_faces
    from ladybug_rhino.grasshopper import all_required_inputs, hide_output, \
        recommended_processor_count
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _run:
    # set the defaults and process all of the inputs
    workers = _cpu_count_ if _cpu_count_ is not None else recommended_processor_count()
    if north_ is not None:  # process the north_
        try:
            north_ = math.degrees(
                to_vector2d(north_).angle_clockwise(Vector2D(0, 1)))
        except AttributeError:  # north angle instead of vector
            north_ = float(north_)
    else:
        north_ = 0
    assert isinstance(_temperature, HourlyDiscontinuousCollection), \
        'Connected _temperature is not a Hourly Data Collection. Got {}'.format(
            type(_temperature))
    assert _temperature.header.unit == 'C', \
        'Connected _temperature must be in Celsius. Got {}.'.format(
            _temperature.header.unit)
    if _timestep_ and _timestep_ != _temperature.header.analysis_period.timestep:
        if _timestep_ < _temperature.header.analysis_period.timestep:
            data = _temperature.cull_to_timestep(_timestep_)
        else:
            data = _temperature.interpolate_to_timestep(_timestep_)
    else:
        data = _temperature
    t_step_per_day = data.header.analysis_period.timestep * 24
    up_thresh = 26 if _up_threshold_ is None else _up_threshold_
    low_thresh = 9 if _low_threshold_ is None else _low_threshold_
    assert up_thresh > low_thresh, 'Input _up_threshold_ [{}] must be greater ' \
        'than input _low_threshold_ [{}].'.format(up_thresh, low_thresh)

    # initialize sunpath based on location and get all of the vectors
    sp = Sunpath.from_location(_location, north_)
    lb_vecs, relevant_temps = [], []
    for dt, temp in zip(data.datetimes, data.values):
        if temp > up_thresh or temp < low_thresh:
            sun = sp.calculate_sun_from_date_time(dt)
            if sun.is_during_day:
                lb_vecs.append(sun.sun_vector_reversed)
                relevant_temps.append(temp)
    vectors = [from_vector3d(lb_vec) for lb_vec in lb_vecs]

    # create the gridded mesh from the geometry
    analysis_mesh = to_joined_gridded_mesh3d(_shade_geo, _grid_size)
    mesh = from_mesh3d(analysis_mesh)
    study_mesh = to_joined_gridded_mesh3d(_study_region, _grid_size / 2)
    points = [from_point3d(pt) for pt in study_mesh.face_centroids]
    hide_output(ghenv.Component, 2)

    # create a series of rays that represent the sun projected through the shade
    int_rays = generate_intersection_rays(points, vectors)

    # if there is context, remove any rays that are blocked by the context
    shade_mesh = join_geometry_to_mesh(context_) \
        if len(context_) != 0 and context_[0] is not None else None

    # intersect the sun rays with the shade mesh
    face_int = intersect_rays_with_mesh_faces(
        mesh, int_rays, shade_mesh, cpu_count=workers)

    # loop through the face intersection result and evaluate the benefit
    pt_div = 1 / float(len(points))
    shade_help, shade_harm, shade_net = [], [], []
    for face_res, face_area in zip(face_int, analysis_mesh.face_areas):
        f_help, f_harm = 0, 0
        for t_ind in face_res:
            t_val = relevant_temps[t_ind]
            if t_val > up_thresh:
                f_help += t_val - up_thresh
            elif t_val < low_thresh:
                f_harm += t_val - low_thresh
        # Normalize by the area of the cell so there's is a consistent metric
        # between cells of different areas.
        # Also, divide the value by t_step_per_day such that the final unit is in
        # degree-days/model unit instead of degree-timesteps/model unit.
        shd_help = ((f_help / face_area) / t_step_per_day) * pt_div
        shd_harm = ((f_harm / face_area) / t_step_per_day) * pt_div
        shade_help.append(shd_help)
        shade_harm.append(shd_harm)
        shade_net.append(shd_help + shd_harm)

    # create the mesh and legend outputs
    graphic = GraphicContainer(shade_net, analysis_mesh.min, analysis_mesh.max, legend_par_)
    graphic.legend_parameters.title = 'degC-days/{}2'.format(units_abbreviation())
    if legend_par_ is None or legend_par_.are_colors_default:
        graphic.legend_parameters.colors = reversed(Colorset.shade_benefit_harm())
    if legend_par_ is None or legend_par_.min is None or legend_par_.max is None:
        bnd_val = max(max(shade_net), abs(min(shade_net)))
        if legend_par_ is None or legend_par_.min is None:
            graphic.legend_parameters.min = -bnd_val
        if legend_par_ is None or legend_par_.max is None:
            graphic.legend_parameters.max = bnd_val
    title = text_objects('Thermal Shade Benefit', graphic.lower_title_location,
                         graphic.legend_parameters.text_height * 1.5,
                         graphic.legend_parameters.font)

    # create all of the visual outputs
    analysis_mesh.colors = graphic.value_colors
    mesh = from_mesh3d(analysis_mesh)
    legend = legend_objects(graphic.legend)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Convert a DataCollection of time-aggregated values to time rate of change units.
_
For example, if the collection has an Energy data type in kWh, this method will
return a collection with an Power data type in W.
-

    Args:
        _data: A houry, sub-hourly or daily data collection that can converted to
            a time rate of change metric. (eg. a data collection of Energy
            values in kWh).

    Returns:
        data_rate: The data collection converted to time rate of changevalues.
            (eg. a data collection of Energy values in kWh).
"""

ghenv.Component.Name = 'LB Time Rate of Change'
ghenv.Component.NickName = 'Rate'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug.datacollection import HourlyDiscontinuousCollection, DailyCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    accept = (HourlyDiscontinuousCollection, DailyCollection)
    assert isinstance(_data, accept), '_data must be a an hourly ot daily data ' \
        'collection. Got {}.'.format(type(_data))
    data_rate = _data.to_time_rate_of_change()



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Calculate wind speed at a specific height above the ground for a given terrain type.
_
By default, the component will calculate wind speed at a height of 1 meter, which
is suitable for most thermal comfort models like PET and SET. Alternatively, by
hooking up the output wind data to the "LB Wind rose" component, a wind rose
for any terrain or at height above the ground can be produced.
_
This component uses the same wind profile function as used by the "LB Wind Profile"
component.
-

    Args:
        _met_wind_vel: A data collection of meteorological wind speed measured at
            the _met_height_ with the _met_terrian [m/s]. Typically, this comes
            from the "LB Import EPW" component. This can also be a number for
            the meteorological wind speed in m/s.
        _height_: The height above the ground to be evaluated in meters. (Default: 1 meter,
            which is suitable for most thermal comfort models like PET and SET.).
        _terrain_: Text string that sets the terrain class associated with the output air_speed.
            This can also be an integer that codes for the terrain. (Default: city).
            Must be one the following.
            _
            0 = city - 50% of buildings above 21m over a distance of at least 2000m upwind.
            1 = suburban - suburbs, wooded areas.
            2 = country - open, with scattered objects generally less than 10m high.
            3 = water - flat areas downwind of a large water body (max 500m inland).
        _met_height_: A number for the height above the ground at which the meteorological
            wind speed is measured in meters. (Default: 10 meters, which is the
            standard used by most airports and EPW files).
        _met_terrain_: Text string that sets the terrain class associated with the
            meteorological wind speed. This can also be an integer that codes
            for the terrain. (Default: country, which is typical of most
            airports where wind measurements are taken). Must be one the following.
            _
            0 = city - 50% of buildings above 21m over a distance of at least 2000m upwind.
            1 = suburban - suburbs, wooded areas.
            2 = country - open, with scattered objects generally less than 10m high.
            3 = water - flat areas downwind of a large water body (max 500m inland).
        log_law_: A boolean to note whether the wind profile should use a logarithmic
            law to determine wind speeds instead of the default power law,
            which is used by EnergyPlus. (Default: False).

    Returns:
        report: Reports, errors, warnings, etc.
        air_speed: A data collection or single value for the air speed at the input
            _height_ above the ground for the input _terrain_. This can be plugged
            into thermal comfort models like PET or SET/PMV. Alternatively, by
            connecting the wind data to the "LB Wind rose" component, a wind rose
            for the input _terrain_ and _height_ above the ground can be produced.
"""

ghenv.Component.Name = 'LB Wind Speed'
ghenv.Component.NickName = 'WindSpeed'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '4'

try:
    from ladybug.datacollection import BaseCollection
    from ladybug.windprofile import WindProfile
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

# dictionary to map integers to terrain types
TERRAIN_TYPES = {
    '0': 'city',
    '1': 'suburban',
    '2': 'country',
    '3': 'water',
    'city': 'city',
    'suburban': 'suburban',
    'country': 'country',
    'water': 'water'
}


if all_required_inputs(ghenv.Component):
    # set default values
    _height_ = 1 if _height_ is None else _height_
    _terrain_ = 'city' if _terrain_ is None else TERRAIN_TYPES[_terrain_.lower()]
    _met_height_ = 10 if _met_height_ is None else _met_height_
    _met_terrain_ = 'country' if _met_terrain_ is None \
        else TERRAIN_TYPES[_met_terrain_.lower()]
    log_law_ = False if log_law_ is None else log_law_

    # create the wind profile object and extract the air speeds
    profile = WindProfile(_terrain_, _met_terrain_, _met_height_, log_law_)
    if isinstance(_met_wind_vel, BaseCollection):
        air_speed = profile.calculate_wind_data(_met_wind_vel, _height_)
    else:  # assume that it is a single number
        _met_wind_vel = float(_met_wind_vel)
        air_speed = profile.calculate_wind(_met_wind_vel, _height_)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Open EPWmap in a web browser.
-

    Args:
        _epw_map: Set to "True" to open EPWmap in a supported browser.
    Returns:
        report: ...
"""

ghenv.Component.Name = 'LB EPWmap'
ghenv.Component.NickName = 'EPWMap'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '0 :: Import'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

# import core Python dependencies
import webbrowser as wb
import os

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

# dictonary of accetable browsers and their default file paths.
acceptable_browsers = [
    ['chrome', 'C:\\Program Files\\Google\\Chrome\\Application\\chrome.exe'],
    ['chrome', 'C:\\Program Files (x86)\\Google\\Chrome\\Application\\chrome.exe'],
    ['firefox', 'C:\\Program Files\\Mozilla Firefox\\firefox.exe'],
    ['chrome', '/Applications/Google Chrome.app/Contents/MacOS/Google Chrome'] # MacOS
]

# URL to epwmap.
url = 'http://www.ladybug.tools/epwmap/'


# function for opening a browser page on Mac
def mac_open(url):
    os.system("open \"\" " + url)


if all_required_inputs(ghenv.Component) and _epw_map is True:
    broswer_found = False
    for browser in acceptable_browsers:
        browser_path = browser[1]
        if broswer_found == False and os.path.isfile(browser_path) == True:
            broswer_found = True
            wb.register(browser[0],  None, wb.BackgroundBrowser(browser_path), 1)
            try:
                wb.get(browser[0]).open(url, 2, True)
                print('Opening epwmap.')
            except ValueError:
                mac_open(url)
    if broswer_found == False:
        print(
            'An accepable broswer was not found on your machine.\n'
            'The default browser will be used but epwmap may not display '
            'correctly there.'
        )
        try:
            wb.open(url, 2, True)
        except ValueError:
            mac_open(url)
else:
    print('Set _epw_map to True.')


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Open a new viewport in Rhino that shows the parallel-projected view from the sun.
_
This is useful for understanding what parts of Rhino geometry are shaded at a
particular hour of the day.
-

    Args:
        _direction: A vector for the direction that the viewport camera faces.
        _position_: A point for the position of the vieport camera in 3D space.
            If no point is provided, the Rhino origin will be used (0, 0, 0).
        look_around_: Optional 2D point (aka. UV coordinates) to tilt the viewport
            camera off from from the input _direction. Values for UV
            coordinates must be between 0 and 1 and these correspond to a
            tilt of 90 degrees in either direction (with 0.5, 0.5 being
            centered on the _direction). Inputting a native Grasshopper
            Slider MD component will allow the most control of view offsetting.
        width_: An optional interger for the width (in pixels) of the Rhino
            viewport that will be generated.
        height_: An optional interger for the height (in pixels) of the Rhino
            viewport that will be generated.
        lens_len_: An optional number that sets the lens length of the viewport
            camera in mm. Typical values are around 20-50mm but wider angle
            views can be achieved by lowering this number to 10 or less.
            If unspecified, the lens length of the currently active Rhino
            viewport will be used.
        mode_: An optional text input for the display mode of the Rhino viewport
            that will be generated. For example: Wireframe, Shaded, Rendered, etc.
            If unspecified, the mode of the currenlty active Rhino viewport
            will be used

    Returns:
        report: The name of the viewport that was opened.
"""

ghenv.Component.Name = 'LB Set View'
ghenv.Component.NickName = 'SetView'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

import math

try:
    from ladybug_geometry.geometry3d import Vector3D
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:
    from ladybug_rhino.togeometry import to_vector3d, to_point2d
    from ladybug_rhino.fromgeometry import from_vector3d
    from ladybug_rhino.viewport import open_viewport, viewport_by_name, \
        set_view_direction, set_view_display_mode
    from ladybug_rhino.grasshopper import all_required_inputs, component_guid, \
        get_sticky_variable, set_sticky_variable
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # get the name of the view and the previous width/height
    view_name = 'SetView_{}'.format(component_guid(ghenv.Component))
    print(view_name)  # print so that the user has the name if needed
    vw = get_sticky_variable('set_view_width')
    vh = get_sticky_variable('set_view_height')

    # if there are look-around coordinates, rotate the direction
    if look_around_ is not None:
        uv_pt = to_point2d(look_around_)
        dir_vec = to_vector3d(_direction)
        v = (uv_pt.y - 0.5) * math.pi
        dir_vec = dir_vec.rotate(dir_vec.cross(Vector3D(0, 0, 1)), v)
        u = -(uv_pt.x - 0.5) * math.pi
        dir_vec = dir_vec.rotate_xy(u)
        _direction = from_vector3d(dir_vec)

    # get the viewport from which the direction will be set
    view_port = None
    if width_ == vw and height_ == vh:  # no need to generate new view; get existing one
        try:
            view_port = viewport_by_name(view_name)
        except ValueError:  # the viewport does not yet exist
            pass
    if view_port is None:
        view_port = open_viewport(view_name, width_, height_)
    set_sticky_variable('set_view_width', width_)
    set_sticky_variable('set_view_height', height_)

    # set the direction of the viewport camera
    set_view_direction(view_port, _direction, _position_, lens_len_)

    # set the display mode if requested
    if mode_:
        set_view_display_mode(view_port, mode_)
